@@ -1,25 +1,34 @@
+// Copyright (c) 2022, Muhammad Asad (masadcv@gmail.com)
+// All rights reserved.
+
+// Redistribution and use in source and binary forms, with or without
+// modification, are permitted provided that the following conditions are met:
+
+// 1. Redistributions of source code must retain the above copyright notice, this
+//    list of conditions and the following disclaimer.
+
+// 2. Redistributions in binary form must reproduce the above copyright notice,
+//    this list of conditions and the following disclaimer in the documentation
+//    and/or other materials provided with the distribution.
+
+// 3. Neither the name of the copyright holder nor the names of its
+//    contributors may be used to endorse or promote products derived from
+//    this software without specific prior written permission.
+
+// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
+// AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
+// IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
+// DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
+// FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
+// DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
+// SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
+// CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
+// OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
+// OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+
 #include <torch/extension.h>
 #include <vector>
-// #include <iostream>
-#include "maxflow-v3.04.src/graph.h"
-
-// void print_shape(torch::Tensor data)
-// {
-//     auto num_dims = data.dim();
-//     std::cout << "Shape: (";
-//     for (int dim = 0; dim < num_dims; dim++)
-//     {
-//         std::cout << data.size(dim);
-//         if (dim != num_dims - 1)
-//         {
-//             std::cout << ", ";
-//         }
-//         else
-//         {
-//             std::cout << ")" << std::endl;
-//         }
-//     }
-// }
+#include "graphcut.h"
 
 // float l1distance(const float &in1, const float &in2)
 // {
@@ -62,18 +71,41 @@ float l2distance(const std::vector<float> &in1, const std::vector<float> &in2)
     return std::sqrt(ret_sum);
 }
 
-torch::Tensor maxflow2d_cpu(const torch::Tensor &image, const torch::Tensor &prob, const float &lambda, const float &sigma)
+torch::Tensor maxflow2d_cpu(const torch::Tensor &image, const torch::Tensor &prob, const float &lambda, const float &sigma, const int &connectivity)
 {
     // get input dimensions
+    const int batch = image.size(0);
     const int channel = image.size(1);
     const int height = image.size(2);
     const int width = image.size(3);
 
-    const int Xoff[2] = {-1, 0};
-    const int Yoff[2] = {0, -1};
+    std::vector<int> Xoff, Yoff;
+    int offsetLen;
 
+    if (connectivity == 0)
+    {
+        std::cout << "torchmaxflow: warning no connectivity provided, falling back to default 4 connectivity" << std::endl;
+    }
+
+    if ((connectivity == 4) || (connectivity == 0)) 
+    {
+        Xoff = {-1,  0};
+        Yoff =  {0, -1};
+        offsetLen = 2;
+    }
+    else if (connectivity == 8) 
+    {
+        Xoff = {-1,  0, -1};
+        Yoff = { 0, -1, -1};
+        offsetLen = 3;
+    }
+    else 
+    {
+        throw std::runtime_error(
+            "torchmaxflow only supports 4 or 8 connectivity for 2D spatial inputs, received connectivity = " + std::to_string(connectivity) + ".");
+    };
     // prepare output
-    torch::Tensor label = torch::zeros_like(image);
+    torch::Tensor label = torch::zeros({batch, 1, height, width}, image.dtype());
 
     // get data accessors
     auto label_ptr = label.accessor<float, 4>();
@@ -81,13 +113,9 @@ torch::Tensor maxflow2d_cpu(const torch::Tensor &image, const torch::Tensor &pro
     auto prob_ptr = prob.accessor<float, 4>();
 
     // prepare graph
-    typedef Graph<float, float, float> GraphType;
     // initialise with graph(num of nodes, num of edges)
-    GraphType *g = new GraphType(height * width, 2 * height * width);
-    g->add_node(height * width);
+    GCGraph<float> g(height * width, 2 * height * width);
 
-    // define max weight value
-    float max_weight = -100000;
     float pval, qval, l2dis, n_weight, s_weight, t_weight, prob_bg, prob_fg;
     int pIndex, qIndex;
     std::vector<float> pval_v(channel);
@@ -97,7 +125,22 @@ torch::Tensor maxflow2d_cpu(const torch::Tensor &image, const torch::Tensor &pro
     {
         for (int w = 0; w < width; w++)
         {
+            pIndex = g.addVtx();
 
+            // avoid log(0)
+            prob_bg = std::max(prob_ptr[0][0][h][w], std::numeric_limits<float>::epsilon());
+            prob_fg = std::max(prob_ptr[0][1][h][w], std::numeric_limits<float>::epsilon());
+            s_weight = -log(prob_bg);
+            t_weight = -log(prob_fg);
+
+            g.addTermWeights(pIndex, s_weight, t_weight);
+        }
+    }
+
+    for (int h = 0; h < height; h++)
+    {
+        for (int w = 0; w < width; w++)
+        {
             if (channel == 1)
             {
                 pval = image_ptr[0][0][h][w];
@@ -109,13 +152,17 @@ torch::Tensor maxflow2d_cpu(const torch::Tensor &image, const torch::Tensor &pro
                     pval_v[c_i] = image_ptr[0][c_i][h][w];
                 }
             }
+
             pIndex = h * width + w;
-            for (int i = 0; i < 2; i++)
+            for (int i = 0; i < offsetLen; i++)
             {
                 const int hn = h + Xoff[i];
                 const int wn = w + Yoff[i];
+                
                 if (hn < 0 || wn < 0)
+                {
                     continue;
+                }
 
                 if (channel == 1)
                 {
@@ -130,35 +177,16 @@ torch::Tensor maxflow2d_cpu(const torch::Tensor &image, const torch::Tensor &pro
                     }
                     l2dis = l2distance(pval_v, qval_v);
                 }
+                l2dis /= sqrt(Xoff[i] * Xoff[i] + Yoff[i] * Yoff[i]);
                 n_weight = lambda * exp(-(l2dis * l2dis) / (2 * sigma * sigma));
                 qIndex = hn * width + wn;
-                g->add_edge(qIndex, pIndex, n_weight, n_weight);
-
-                if (n_weight > max_weight)
-                {
-                    max_weight = n_weight;
-                }
+                g.addEdges(qIndex, pIndex, n_weight, n_weight);
             }
         }
     }
 
-    max_weight = 1000 * max_weight;
-    for (int h = 0; h < height; h++)
-    {
-        for (int w = 0; w < width; w++)
-        {
-
-            // avoid log(0)
-            prob_bg = std::max(prob_ptr[0][0][h][w], std::numeric_limits<float>::epsilon());
-            prob_fg = std::max(prob_ptr[0][1][h][w], std::numeric_limits<float>::epsilon());
-            s_weight = -log(prob_bg);
-            t_weight = -log(prob_fg);
-
-            pIndex = h * width + w;
-            g->add_tweights(pIndex, s_weight, t_weight);
-        }
-    }
-    double flow = g->maxflow();
+    g.maxFlow();
+    // float flow = g.maxFlow();
     // std::cout << "max flow: " << flow << std::endl;
 
     int idx = 0;
@@ -166,28 +194,60 @@ torch::Tensor maxflow2d_cpu(const torch::Tensor &image, const torch::Tensor &pro
     {
         for (int w = 0; w < width; w++)
         {
-            label_ptr[0][0][h][w] = 1 - g->what_segment(idx);
+            label_ptr[0][0][h][w] = g.inSourceSegment(idx) ? 1.0 : 0.0;
             idx++;
         }
     }
-    delete g;
     return label;
 }
 
-torch::Tensor maxflow3d_cpu(const torch::Tensor &image, const torch::Tensor &prob, const float &lambda, const float &sigma)
+torch::Tensor maxflow3d_cpu(const torch::Tensor &image, const torch::Tensor &prob, const float &lambda, const float &sigma, const int &connectivity)
 {
     // get input dimensions
+    const int batch = image.size(0);
     const int channel = image.size(1);
     const int depth = image.size(2);
     const int height = image.size(3);
     const int width = image.size(4);
 
-    const int Xoff[3] = {-1, 0, 0};
-    const int Yoff[3] = {0, -1, 0};
-    const int Zoff[3] = {0, 0, -1};
+    std::vector<int> Xoff, Yoff, Zoff;
+    int offsetLen;
+        
+    if(connectivity == 0)
+    {
+        // no connectivity provided, issue warning and use default connectivity
+        std::cout << "torchmaxflow: warning no connectivity provided, falling back to default 6 connectivity" << std::endl;
+    }
+
+    if ((connectivity == 6) || (connectivity == 0)) 
+    {
+        Xoff = {-1,  0,  0};
+        Yoff = { 0, -1,  0};
+        Zoff = { 0,  0, -1};
+        offsetLen = 3;
+    }
+    else if (connectivity == 18) 
+    {
+        Xoff = {-1,  0,  0, -1, -1,  0};
+        Yoff = { 0, -1,  0, -1,  0, -1};
+        Zoff = { 0,  0, -1,  0, -1, -1};
+        offsetLen = 6;
+    }
+    else if (connectivity == 26) 
+    {
+        Xoff = {-1,  0,  0, -1, -1,  0, -1};
+        Yoff = { 0, -1,  0, -1,  0, -1, -1};
+        Zoff = { 0,  0, -1,  0, -1, -1, -1};
+        offsetLen = 7;
+    }
+    else 
+    {
+        throw std::runtime_error(
+            "torchmaxflow only supports 6, 18 or 26 connectivity for 3D spatial inputs, received connectivity = " + std::to_string(connectivity) + ".");
+    };
 
     // prepare output
-    torch::Tensor label = torch::zeros_like(image);
+    torch::Tensor label = torch::zeros({batch, 1, depth, height, width}, image.dtype());
 
     // get data accessors
     auto label_ptr = label.accessor<float, 5>();
@@ -195,17 +255,31 @@ torch::Tensor maxflow3d_cpu(const torch::Tensor &image, const torch::Tensor &pro
     auto prob_ptr = prob.accessor<float, 5>();
 
     // prepare graph
-    typedef Graph<float, float, float> GraphType;
     // initialise with graph(num of nodes, num of edges)
-    GraphType *g = new GraphType(depth * height * width, 2 * depth * height * width);
-    g->add_node(depth * height * width);
+    GCGraph<float> g(depth * height * width, 2 * depth * height * width);
 
-    // define max weight value
-    float max_weight = -100000;
     float pval, qval, l2dis, n_weight, s_weight, t_weight, prob_bg, prob_fg;
     int pIndex, qIndex;
     std::vector<float> pval_v(channel);
     std::vector<float> qval_v(channel);
+    for (int d = 0; d < depth; d++)
+    {
+        for (int h = 0; h < height; h++)
+        {
+            for (int w = 0; w < width; w++)
+            {
+                pIndex = g.addVtx();
+
+                // avoid log(0)
+                prob_bg = std::max(prob_ptr[0][0][d][h][w], std::numeric_limits<float>::epsilon());
+                prob_fg = std::max(prob_ptr[0][1][d][h][w], std::numeric_limits<float>::epsilon());
+                s_weight = -log(prob_bg);
+                t_weight = -log(prob_fg);
+
+                g.addTermWeights(pIndex, s_weight, t_weight);
+            }
+        }
+    }
 
     for (int d = 0; d < depth; d++)
     {
@@ -224,15 +298,18 @@ torch::Tensor maxflow3d_cpu(const torch::Tensor &image, const torch::Tensor &pro
                         pval_v[c_i] = image_ptr[0][c_i][d][h][w];
                     }
                 }
-                pIndex = d * height * width + h * width + w;
 
-                for (int i = 0; i < 3; i++)
+                pIndex = d * height * width + h * width + w;
+                for (int i = 0; i < offsetLen; i++)
                 {
                     const int dn = d + Xoff[i];
                     const int hn = h + Yoff[i];
                     const int wn = w + Zoff[i];
+
                     if (dn < 0 || hn < 0 || wn < 0)
+                    {
                         continue;
+                    }
 
                     if (channel == 1)
                     {
@@ -245,41 +322,21 @@ torch::Tensor maxflow3d_cpu(const torch::Tensor &image, const torch::Tensor &pro
                         {
                             qval_v[c_i] = image_ptr[0][c_i][dn][hn][wn];
                         }
-                        l2dis = l2distance(pval_v, qval_v);
+                        l2dis = l2distance(pval_v, qval_v); 
                     }
+                    l2dis /= sqrt(Xoff[i] * Xoff[i] + Yoff[i] * Yoff[i] + Zoff[i] * Zoff[i]);
                     n_weight = lambda * exp(-(l2dis * l2dis) / (2 * sigma * sigma));
                     qIndex = dn * height * width + hn * width + wn;
-                    g->add_edge(qIndex, pIndex, n_weight, n_weight);
-
-                    if (n_weight > max_weight)
-                    {
-                        max_weight = n_weight;
-                    }
+                    g.addEdges(qIndex, pIndex, n_weight, n_weight);
                 }
             }
         }
     }
 
-    max_weight = 1000 * max_weight;
-    for (int d = 0; d < depth; d++)
-    {
-        for (int h = 0; h < height; h++)
-        {
-            for (int w = 0; w < width; w++)
-            {
-                // avoid log(0)
-                prob_bg = std::max(prob_ptr[0][0][d][h][w], std::numeric_limits<float>::epsilon());
-                prob_fg = std::max(prob_ptr[0][1][d][h][w], std::numeric_limits<float>::epsilon());
-                s_weight = -log(prob_bg);
-                t_weight = -log(prob_fg);
-
-                pIndex = d * height * width + h * width + w;
-                g->add_tweights(pIndex, s_weight, t_weight);
-            }
-        }
-    }
-    double flow = g->maxflow();
+    g.maxFlow();
+    // float flow = g.maxFlow();
     // std::cout << "max flow: " << flow << std::endl;
+
     int idx = 0;
     for (int d = 0; d < depth; d++)
     {
@@ -287,12 +344,11 @@ torch::Tensor maxflow3d_cpu(const torch::Tensor &image, const torch::Tensor &pro
         {
             for (int w = 0; w < width; w++)
             {
-                label_ptr[0][0][d][h][w] = 1 - g->what_segment(idx);
+                label_ptr[0][0][d][h][w] = g.inSourceSegment(idx) ? 1.0 : 0.0;
                 idx++;
             }
         }
     }
-    delete g;
     return label;
 }
 
@@ -303,7 +359,6 @@ void add_interactive_seeds(torch::Tensor &prob, const torch::Tensor &seed, const
     //  "Interactive medical image segmentation using deep learning with image-specific fine tuning."
     //  IEEE TMI (2018).
 
-    const int channel = prob.size(1);
     if (num_dims == 4) // 2D
     {
         const int height = prob.size(2);
